import numpy
import scipy.sparse.linalg
import arcview
import arcpy
import copy
import os
from scipy.sparse import dok_matrix
from scipy.sparse.linalg import lsmr
from scipy.sparse.linalg import spsolve
from scipy.sparse.linalg import LinearOperator

#
# Machinery for Laplacian infill
#

def pefinfill_laplacian(data) :
    """convolution with a 2D Laplacian filter operator"""
    result = -6.0*data.copy()
    for i in [-1, 0, 1] :
        result += numpy.roll(data,i,axis=0)
    for j in [-1, 0, 1] :
        result += numpy.roll(data,j,axis=1)
    return -0.25*result

class LaplacianInterpLinearOperator(scipy.sparse.linalg.LinearOperator) :
    def __init__(self,subBox,polygonInterior,noDataValue=None) :
        """ Missing data Laplacian smooth infill
            subBox contains polygon and missing data are inside polygon
        """
        self._subBox = subBox.copy()
        self._polygonInterior = polygonInterior.copy()
        self._noDataValue = noDataValue
        self._datalocs = (self._subBox != noDataValue)
        self._missinglocs = numpy.logical_and(self._subBox == self._noDataValue, polygonInterior != 0)
#       self._missinglocs =  self._subBox == self._noDataValue
        self._nMissing = numpy.count_nonzero(self._missinglocs)
        super(PefInterpLinearOperator,self).__init__(dtype=self._subBox.dtype,shape=(self._nMissing,self._nMissing))

    def get_datalocs(self):
        return self._datalocs

    def get_missinglocs(self):
        return self._missinglocs

    def get_nMissing(self):
        return self._nMissing

    def __repr__(self):
        return "{0}({1},{2},{3},{4})".format(self.__name__,_subBox.repr(), \
               _polygonInterior.repr(), \
               _noDataValue.repr())

    def __str__(self):
        return ("{0}(bounding box,indices of polygon interior points," + \
               "noDataValue=None)").format( \
               self.__name__)

    def _matvec(self,v) :
        temp01 = numpy.zeros(self._subBox.shape,dtype=numpy.float64)
        temp01[self._missinglocs] = v[:]
        w = pefinfill_laplacian(self._opDim,temp01)
        return w[self._missinglocs]

    def _rmatvec(self,w) :
        temp02 = numpy.zeros(self._subBox.shape,dtype=numpy.float64)
        temp02[self._missinglocs] = w[:]
        v = pefinfill_laplacian(self._opDim,temp02)
        return v[self._missinglocs]

#
# Machinery for PEF interpolation
#

def pefinfill_helix_conv(opDim,filter,data) :
    """Helix convolution with an opDim x opDim filter operator"""
    result = numpy.zeros(data.shape,dtype=numpy.float64)
    ny = data.shape[0]
    jmin = 0
    fmin = 0
    for i in range(0, opDim) :
        temp00b = numpy.roll(data,i,axis=0)
        for j in range(jmin, opDim) :
            temp00c = numpy.roll(temp00b,j,axis=1)
            result += filter[fmin]*temp00c
            fmin += 1
        jmin = 1-opDim
    return result
def pefinfill_helix_conv(opDim,filter,data) :
    """Helix convolution with an opDim x opDim filter operator"""
    result = numpy.zeros(data.shape,dtype=numpy.float64)
    ny = data.shape[0]
    jmin = 0
    fmin = 0
    for i in range(0, opDim) :
        temp00b = numpy.roll(data,i,axis=0)
        for j in range(jmin, opDim) :
            temp00c = numpy.roll(temp00b,j,axis=1)
            result += filter[fmin]*temp00c
            fmin += 1
        jmin = 1-opDim
    return result

def pefinfill_helix_corr(opDim,filter,data) :
    """Helix correlation with an opDim x opDim filter operator"""
    result = numpy.zeros(data.shape,dtype=numpy.float64)
    ny = data.shape[0]
    jmin = 0
    fmin = 0
    for i in range(0, opDim) :
        temp00b = numpy.roll(data,-i,axis=0)
        for j in range(jmin, opDim) :
            temp00c = numpy.roll(temp00b,-j,axis=1)
            result += filter[fmin]*temp00c
            fmin += 1
        jmin = 1-opDim
    return result

def pef_nst_pass(opDim,filter,data,dataAbsSquareMax,missMask,messages) :
    """Nonstationary prediction error missing data infill pass"""
    """From Claerbout and Wang 2018"""
#   fidi = open("pef_nst_input","ab")
#   numpy.float32(data).tofile(fidi)
#   fidi.close()
#   fido = open("pef_nst_output","ab")
#   fidr = open("pef_nst_resid","ab")
    epsilonY = 0.1
    r = numpy.zeros(data.shape,dtype=numpy.float64)
#    d2 = numpy.zeros(data.shape,dtype=numpy.float64)
    d2 = numpy.full(data.shape,2.0*opDim*opDim*dataAbsSquareMax)
    nx = data.shape[1]
    ny = data.shape[0]
    messages.addMessage("data.shape="+str(data.shape))
    y = data.copy()
    a = -filter.copy()
    for iy in range(opDim,ny-opDim) :
        for ix in range(opDim,nx-2*opDim) :
            jmin = 0
            fmin = 0
            for i in range(0, opDim) :
                for j in range(jmin, opDim) :
#                   d2[iy,ix] += y[(iy-i)%ny,(ix-j)%nx]*y[(iy-i)%ny,(ix-j)%nx]
                    r[iy,ix] += a[fmin]*y[(iy-i)%ny,(ix-j)%nx]
                    fmin += 1
                jmin = 1-opDim
            if d2[iy,ix] > 0.0 :
                epsilonA = 0.05/d2[iy,ix]
            else :
                epsilonA = 0.05
            jmin = 1
            fmin = 1
            for i in range(0, opDim) :
                for j in range(jmin, opDim) :
                    a[fmin] -= epsilonA*r[iy,ix]*y[(iy-i)%ny,(ix-j)%nx]
                    fmin += 1
                jmin = 1-opDim
            jmin = 0
            fmin = 0
            for i in range(0, opDim) :
                for j in range(jmin, opDim) :
                    if missMask[(iy-i)%ny,(ix-j)%nx] :
#                       messages.addMessage("miss("+str(iy)+","+str(ix)+") a["+str(fmin)+"]="+str(a[fmin]))
                        y[(iy-i)%ny,(ix-j)%nx] -= epsilonY*r[iy,ix]*a[fmin]
                    fmin += 1
                jmin = 1-opDim
#   numpy.float32(r).tofile(fidr)
#   numpy.float32(y).tofile(fido)
#   fidr.close()
#   fido.close()
    return y

def box_smooth(halfwidth,data) :
    """square box smoothing"""
    fastax = len(data.shape)-1
    slowax = fastax-1
    result = numpy.zeros(data.shape,dtype=data.dtype)
    for ix in range(1-halfwidth,halfwidth) :
        temp0 = numpy.roll(data,ix,axis=fastax)
        for iy in range(1-halfwidth,halfwidth) :
            result += numpy.roll(temp0,iy,axis=slowax)
    return result/((2*halfwidth-1)*(2*halfwidth-1))

def focal_infill(data,no_data_value,messages) :
    """progressive box averaging to infill no_data_val areas"""
    doit = True
    fastax = len(data.shape)-1
    slowax = fastax-1
    r0 = data.copy()
    iterno = 0
    messages.addmessage("data.shape="+str(data.shape))
    while True :
        deadlocs = (r0 == no_data_value)
        messages.addMessage("numdead=" + str(numpy.count_nonzero(deadlocs)))
        if not numpy.any(deadlocs) :
            break
        livecount = numpy.zeros(data.shape,dtype=int)
        sumvals = numpy.zeros(data.shape,dtype=data.dtype)
        r1 = numpy.roll(r0,-1,axis=fastax)
        r2 = numpy.roll(r0, 1,axis=fastax)
        r3 = numpy.roll(r0,-1,axis=slowax)
        r4 = numpy.roll(r0, 1,axis=slowax)
        r5 = numpy.roll(r1,-1,axis=slowax)
        r6 = numpy.roll(r1, 1,axis=slowax)
        r7 = numpy.roll(r2,-1,axis=slowax)
        r8 = numpy.roll(r2, 1,axis=slowax)
#       deadlocs.astype(numpy.float32).tofile("infill"+str(iterno))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r1 != no_data_value)))
        numpy.add(sumvals,r1,out=sumvals,where=numpy.logical_and(deadlocs , (r1 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r2 != no_data_value)))
        numpy.add(sumvals,r2,out=sumvals,where=numpy.logical_and(deadlocs , (r2 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r3 != no_data_value)))
        numpy.add(sumvals,r3,out=sumvals,where=numpy.logical_and(deadlocs , (r3 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r4 != no_data_value)))
        numpy.add(sumvals,r4,out=sumvals,where=numpy.logical_and(deadlocs , (r4 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r5 != no_data_value)))
        numpy.add(sumvals,r5,out=sumvals,where=numpy.logical_and(deadlocs , (r5 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r6 != no_data_value)))
        numpy.add(sumvals,r6,out=sumvals,where=numpy.logical_and(deadlocs , (r6 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r7 != no_data_value)))
        numpy.add(sumvals,r7,out=sumvals,where=numpy.logical_and(deadlocs , (r7 != no_data_value)))
        numpy.add(livecount,1,out=livecount,where=numpy.logical_and(deadlocs , (r8 != no_data_value)))
        numpy.add(sumvals,r8,out=sumvals,where=numpy.logical_and(deadlocs , (r8 != no_data_value)))
        numpy.divide(sumvals,livecount,out=livecount,where=numpy.logical_and(deadlocs, (livecount > 0)))
        r0 = numpy.where(sumvals,r0,numpy.logical_and(deadlocs, (livecount > 0)))
        iterno = iterno+1
#
    return r0
       
def laplacian_infill(data,no_data_value,messages) :
    """solve sparse linear system for Laplacian interpolation"""
    """assumes 3d array, with first dim numpy.newaxis as needed"""
    r0 = data.copy()
    messages.addMessage("r0.shape="+str(r0.shape))
    for islice in range(0,r0.shape[0]) :
        temp21 = numpy.zeros(r0.shape[1:],dtype=numpy.float64)
        temp23 = (r0[islice,...] == no_data_value)
        nMissing = numpy.count_nonzero(temp23)
        r0[islice,...][temp23] = 0.0
        temp24 = pefinfill_laplacian(r0[islice,...])
        missRhs = temp24[temp23]
        missRhs *= -1.0
        misstemp1 = numpy.full(r0.shape[1:],-1)
        misstemp1[temp23] = numpy.arange(0,nMissing,1)
        miss_dok_matrix = scipy.sparse.dok_matrix((nMissing,nMissing))
        for miss_i in [-1,1] :
            for iax in [0,1] :
                miss_temp00a = numpy.roll(misstemp1,miss_i,axis=iax)
                miss_temp00b = miss_temp00a[temp23]
                for miss_k in range(0,miss_temp00b.size) :
                    if miss_temp00b[miss_k] > -1 :
                        miss_dok_matrix[miss_k,miss_temp00b[miss_k]] = -0.25
        for miss_j in range(0,missRhs.size) :
            miss_dok_matrix[miss_j,miss_j] = 1.0
        missingData = scipy.sparse.linalg.spsolve(miss_dok_matrix,missRhs)
        r0[islice,...][temp23] = missingData
#       numpy.float32(r0).tofile("infill0")
#
    return r0
       

class PefInterpLinearOperator(scipy.sparse.linalg.LinearOperator) :
    def __init__(self,subBox,polygonInterior,pefFilter,noDataValue=None) :
        """ Missing data infill with prediction error filtering
            subBox contains polygon and missing data are inside polygon
        """
        self._subBox = subBox.copy()
        self._polygonInterior = polygonInterior.copy()
        self._pefFilter = pefFilter.copy()
        self._pefSize = self._pefFilter.size
        self._opDim = int(round(0.5*(1+math.sqrt(2.0*self._pefSize-1))))
        self._noDataValue = noDataValue
        self._datalocs = (self._subBox != noDataValue)
        self._missinglocs = numpy.logical_and(self._subBox == self._noDataValue, polygonInterior != 0)
#       self._missinglocs =  self._subBox == self._noDataValue
        self._nMissing = numpy.count_nonzero(self._missinglocs)
        super(PefInterpLinearOperator,self).__init__(dtype=self._subBox.dtype,shape=(self._nMissing,self._nMissing))

    def get_datalocs(self):
        return self._datalocs

    def get_missinglocs(self):
        return self._missinglocs

    def get_pefSize(self):
        return self._pefSize

    def get_nMissing(self):
        return self._nMissing

    def __repr__(self):
        return "{0}({1},{2},{3},{4})".format(self.__name__,_subBox.repr(), \
               _polygonInterior.repr(),_pefFilter.repr(), \
               _noDataValue.repr())

    def __str__(self):
        return ("{0}(bounding box,indices of polygon interior points," + \
               "prediction error filter, noDataValue=None)").format( \
               self.__name__)

    def _matvec(self,v) :
        temp01 = numpy.zeros(self._subBox.shape,dtype=numpy.float64)
        temp01[self._missinglocs] = v[:]
        w = pefinfill_helix_conv(self._opDim,self._pefFilter,temp01)
        return w[self._missinglocs]

    def _rmatvec(self,w) :
        temp02 = numpy.zeros(self._subBox.shape,dtype=numpy.float64)
        temp02[self._missinglocs] = w[:]
        v = pefinfill_helix_corr(self._opDim,self._pefFilter,temp02)
        return v[self._missinglocs]

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "PefInfill"
        self.alias = ""

        # List of tool classes associated with this toolbox
        self.tools = [PefInfillTool]

class PefInfillTool(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "PEF Grid Infill"
        self.description = "Grid infill using prediction error filtering"
        self.category = "Interpolation"
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        param0 = arcpy.Parameter(
            displayName="Input Raster",
            name="input_raster",
            datatype="GPRasterLayer",
            parameterType="Required",
            direction="Input")

        param1 = arcpy.Parameter(
            displayName="Area of Interest",
            name="area_of_interest",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input",
            multiValue=True)
        param1.filter.list = ["Polygon"]

        param2 = arcpy.Parameter(
            displayName="Output Raster",
            name="output_raster",
            datatype="DERasterDataset",
            parameterType="Required",
            direction="Output")
        param2.parameterDependencies=[param0.name]
        param2.schema.clone = True

        param3 = arcpy.Parameter(
            displayName="Use separate design area",
            name="separate_design",
            datatype="GPBoolean",
            parameterType="Required",
            direction="Input")
        param3.value = False

        param4 = arcpy.Parameter(
            displayName="Design Area",
            name="design_area",
            datatype="GPFeatureLayer",
            parameterType="Optional",
            direction="Input",
            enabled=False,
            multiValue=True)
        param4.filter.list = ["Polygon"]

        params = [param0, param1, param2, param3, param4]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
#       if arcpy.CheckOutExtension("Spatial") == "CheckedOut" :
#           return True
#       else:
#           return False
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        if parameters[3].value :
            parameters[4].parameterType = "Required"
        else :
            parameters[4].parameterType = "Optional"
        if parameters[3].altered :
            parameters[4].enabled = parameters[3].value
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        parmsGood = True
        if not parameters[0].value :
            messages.setErrorMessage("Missing Raster Input")
            parmsGood = False
        if not parameters[1].value :
            messages.setErrorMessage("Missing Polygon Area of Interest Input")
            parmsGood = False
        if not parameters[2].value :
            messages.setErrorMessage("Missing Raster Output")
            parmsGood = False
        if not parmsGood :
            return
        
        messages.addMessage("param0=" + parameters[0].valueAsText)
        input_raster = arcpy.Raster(parameters[0].valueAsText)
#       outFocalStatistics = arcpy.sa.FocalStatistics(input_raster,arcpy.sa.NbrCircle(5,"CELL"),"MEAN","DATA")
#       arcpy.CheckInExtension("Spatial")
        messages.addMessage("input_raster=" + str(input_raster))
        maxop = 3 # maximum operator size 3x5
        maxop = 7 # maximum operator size 7x13
        maxop = 5 # maximum operator size 5x9
        min_statistics_fold = 9 # want at least 9 * no. filter coeffs eqns.
#       desc_ras = arcpy.Describe(parameters[0].valueAsText+"/Band_1")
        desc_ras = arcpy.Describe(parameters[0].value)
        messages.addMessage("desc_ras="+desc_ras.catalogPath)
        messages.addMessage("desc_ras.extent="+str(desc_ras.extent))
        messages.addMessage("desc_ras.meanCellWidth="+str(desc_ras.meanCellWidth))
        messages.addMessage("desc_ras.meanCellHeight="+str(desc_ras.meanCellHeight))
        messages.addMessage("desc_ras.spatialReference="+str(desc_ras.spatialReference))
        messages.addMessage("desc_ras.noDataValue="+str(desc_ras.noDataValue))
        messages.addMessage("param1="+parameters[1].valueAsText)
#        area_of_interest=arcpy.Polygon(parameters[1].value,desc_ras.spatialReference)
        area_of_interest=parameters[1]
#
#       Create 1/0 mask corresponding to polygon
#
        tempras00 = arcpy.RasterToNumPyArray(input_raster,nodata_to_value=desc_ras.noDataValue)
        if tempras00.ndim == 2 :
            tempras0 = tempras00[numpy.newaxis,:,:]
        else :
            tempras0 = tempras00
        liveras0 = (tempras0 != desc_ras.noDataValue)
        deadras0 = (tempras0 == desc_ras.noDataValue)
        temprasSmooth = laplacian_infill(tempras0,desc_ras.noDataValue,messages)
#       numpy.float32(temprasSmooth*liveras0).tofile("infillLaplacian")
        temprasSmooth = 0.0*box_smooth(maxop*1,laplacian_infill(tempras0,desc_ras.noDataValue,messages))
#       temprasSmooth = liveras0*box_smooth(maxop*5,tempras0*liveras0)                         + deadras0*laplacian_infill(tempras0,desc_ras.noDataValue,messages)
#       numpy.float32(temprasSmooth).tofile("infillSmooth")
        tempras0[liveras0] -= temprasSmooth[liveras0]
        minrasval = numpy.nanmin(tempras0[liveras0])
        maxrasval = numpy.nanmax(tempras0[liveras0])
        dataAbsSquareMax = max(minrasval*minrasval,maxrasval*maxrasval)
        meanrasval = numpy.nanmean(tempras0[liveras0])
        stdrasval = numpy.nanstd(tempras0[liveras0])
# subtract mean, divide by variance
        tempras0[liveras0] -= meanrasval
        varscl = stdrasval*stdrasval
        if (varscl == 0.0) :
            varscl = 1.0
        tempras0[liveras0] /= varscl
#       numpy.float32(tempras0).tofile("infill0")
        tempras1 = numpy.ones(tempras0.shape)
#
# We remember where the null values were and make a copy with them
# replaced by zeros to start with.
#
        tempras4 = tempras0.copy()
        tempras5 = tempras0.copy()
        tempras4[deadras0] = 0
#       numpy.float32(tempras4).tofile("infill1")
        messages.addMessage("meanrasval="+str(meanrasval)+" stdrasval="+str(stdrasval))
        sumX = numpy.sum(tempras1[0],axis=0)
        sumY = numpy.sum(tempras1[0],axis=1)
        iy1 = int(numpy.nonzero(sumX)[0][0])
        iy2 = 1 + int(numpy.nonzero(sumX)[0][-1])
        ny  = iy2 - iy1
        nypad = ny+2*maxop
        ix1 = int(numpy.nonzero(sumY)[0][0])
        ix2 = 1 + int(numpy.nonzero(sumY)[0][-1])
        nx  = ix2 - ix1
        nxpad = nx+2*maxop
        tempras1[liveras0] = 0
#
        ijtransp = numpy.array([
                               [ 1, 1,0],[ 1, 1,1],[ 1,-1,0],[ 1,-1,1],
                               [-1, 1,0],[-1, 1,1],[-1,-1,0],[-1,-1,1]
                               ],
                               dtype=int)
#       ijtransp = numpy.array([
#                              [ -1,-1,0],
#                              ],
#                              dtype=int)
        pefDict = {}
        for ijtix in ijtransp :
            idir = ijtix[0]
            jdir = ijtix[1]
            dotransp = ijtix[2] 
            if dotransp :
                 nxtransp = ny
                 nytransp = nx
                 nxpadtransp = nypad
                 nypadtransp = nxpad
            else :
                 nxtransp = nx
                 nytransp = ny
                 nxpadtransp = nxpad
                 nypadtransp = nypad
            working_area = numpy.transpose(tempras4[:,ix1:ix2,iy1:iy2][:,::idir,::jdir],(0,dotransp+1,2-dotransp))
            working_area5 = numpy.transpose(tempras5[:,ix1:ix2,iy1:iy2][:,::idir,::jdir],(0,dotransp+1,2-dotransp))
            polygon_area = numpy.transpose(tempras1[:,ix1:ix2,iy1:iy2][:,::idir,::jdir],(0,dotransp+1,2-dotransp))
            messages.addMessage("working_area.ndim="+str(working_area.ndim))
            messages.addMessage("working_area.shape="+str(working_area.shape))
#
# Now we get to work figuring the largest pefinfill operator that
# fits enough non-noDataValue places in the working area to
# have 9 times as many locations as cells in the operator
#
            mask = numpy.ones(working_area.shape[1:],dtype=numpy.int16)
            mask[numpy.nonzero(working_area5[0,:,:] == desc_ras.noDataValue)] = 0
            mask_pad = numpy.pad(mask,((0,2*maxop),(0,2*maxop)),'constant',constant_values=0)
#    
# should modify so that only a border is added each time
#
            for opDim in range(2,maxop+1) :
                opSize = (opDim*opDim)+(opDim-1)*(opDim-1) # pefinfill operators
                fit = numpy.zeros(mask_pad.shape)
                jmin = 0
                for i in range(0, opDim) :
                    temp00b = numpy.roll(mask_pad,i,axis=0)
                    for j in range(jmin, opDim) :
                        temp00c = numpy.roll(temp00b,j,axis=1)
                        fit += temp00c
                    jmin = 1-opDim
                nStatistics = (fit == opSize).size
                if nStatistics < (9*opSize) :
                    break
#
# back off one size as necessary to get sufficient statistics
#
            messages.addMessage("nStatistics(1)="+str(nStatistics))
            messages.addMessage("opDim(1)="+str(opDim))
            if(nStatistics < (9*opSize)) :
                opDim = opDim - 1
            else :
                opDim = maxop
#
# ensure that the operator is at least 2x2!
#
            if(opDim < 2) :
                continue
#
# recalculate the "live" locations for the pef accumulation
#
            opSize = (opDim*opDim)+(opDim-1)*(opDim-1)
            fit = numpy.zeros(mask_pad.shape)
            jmin = 0
            for i in range(0,opDim) :
                temp00b = numpy.roll(mask_pad,i,axis=0)
                for j in range(jmin,opDim) :
                    temp00c = numpy.roll(temp00b,j,axis=1)
                    fit += temp00c
                jmin = 1-opDim
#
# build explicit matrix and right hand side for call to solver
#
            livelocs = (fit == opSize).nonzero()
            nEqns = numpy.size(livelocs[0])
            nVars = opSize - 1
            messages.addMessage("opDim="+str(opDim)+" opSize="+str(opSize))
            messages.addMessage("nEqns="+str(nEqns)+" nVars="+str(nVars))
#
# loop over number of images in working area
#
            for isheet in range(working_area.shape[0]) :
                data_pad = numpy.pad(working_area[isheet,:,:],((0,2*maxop),(0,2*maxop)),'constant',constant_values=(desc_ras.noDataValue,))
                pefMatrix = numpy.empty( (nEqns,nVars), dtype=numpy.float64 )
                pefRhs = numpy.empty( [nEqns], dtype=numpy.float64)
                jmin = 1
                fmin = 0
                for i in range(0,opDim) :
                    temp00b = numpy.roll(data_pad,i,axis=0)
                    for j in range(jmin,opDim) :
                        temp00c = numpy.roll(temp00b,j,axis=1)
                        pefMatrix[:,fmin] = temp00c[livelocs]
                        fmin += 1
                    jmin = 1-opDim
                pefRhs[:] = data_pad[livelocs]
#
# calculate prediction filter
#
                (predFilter, istop, itn, normr, normar, norma, conda, normx) = \
                       scipy.sparse.linalg.lsmr(pefMatrix,pefRhs,maxiter=nVars*30)
                messages.addMessage("pefcalc: istop="+str(istop)+" itn="+str(itn)+" normr="+str(normr)+" normar="+str(normar)+" norma="+str(norma)+" conda="+str(conda)+" normx="+str(normx))
#               predFilter[:] = 0.0
                pefFilter = numpy.pad(predFilter,((1,0)),'constant',constant_values=((-1.0,)))
                pefDict[(isheet,tuple(ijtix.tolist()))] = ( opDim, pefFilter )
#
#  Now make multiple, multidirection passes over the data to fill in
#  missing data holes.
#
#       fid = open("pef_nst_output","wb")
#       fid.close()
#       fid = open("pef_nst_input","wb")
#       fid.close()
#       fid = open("pef_nst_resid","wb")
#       fid.close()
        niter = 8*64/8
        for iterno in range(0,niter) :
            for ijtix in ijtransp :
                idir = ijtix[0]
                jdir = ijtix[1]
                dotransp = ijtix[2] 
                if dotransp :
                     nxtransp = ny
                     nytransp = nx
                     nxpadtransp = nypad
                     nypadtransp = nxpad
                else :
                     nxtransp = nx
                     nytransp = ny
                     nxpadtransp = nxpad
                     nypadtransp = nypad
                messages.addMessage("iterno "+str(iterno+1)+" (idir,jdir,dotransp) "+str(ijtix))
                working_area = numpy.transpose(tempras4[:,ix1:ix2,iy1:iy2][:,::idir,::jdir],(0,dotransp+1,2-dotransp))
                polygon_area = numpy.transpose(tempras1[:,ix1:ix2,iy1:iy2][:,::idir,::jdir],(0,dotransp+1,2-dotransp))
#
# loop over number of images in working area
#
                for isheet in range(working_area.shape[0]) :
                    data_pad = numpy.pad(working_area[isheet,:,:],((0,2*maxop),(0,2*maxop)),'constant',constant_values=(0.0*desc_ras.noDataValue,))
                    messages.addMessage("one pef_nst_pass start data_pad_[51,51]="+str(data_pad[51,51]))
                    polymask_pad = numpy.pad(polygon_area[isheet,:,:],((0,2*maxop),(0,2*maxop)),'constant',constant_values=(0,))
#
                    ( opdim, pefFilterList ) = pefDict[(isheet,tuple(ijtix.tolist()))]
                    pefFilter = numpy.array(pefFilterList)
                    messages.addMessage("isheet="+str(isheet)+" opdim="+str(opdim))
                    messages.addMessage("ijtix("+str(ijtix)+")")
#                   numpy.float32(polygon_area[isheet,:,:]).tofile("infill2")
                    data_pad_result = pef_nst_pass(opdim,-pefFilter,data_pad,dataAbsSquareMax,polymask_pad,messages)               
#                   numpy.float32(data_pad_result).tofile("infill_"+str(idir)+"_"+str(jdir)+"_"+str(dotransp)+"_"+str(iterno))
                    messages.addMessage("one pef_nst_pass finished data_pad_result[51,51]="+str(data_pad_result[51,51]))
                    working_area[isheet,:,:] = data_pad_result[0:(data_pad.shape[0]-2*maxop),0:(data_pad.shape[1]-2*maxop)]
                    messages.addMessage("working_area[0,51,51]="+str(working_area[0,51,51]))
                    tempras4[isheet,ix1:ix2,iy1:iy2] = numpy.transpose(working_area[isheet,:,:],(dotransp,1-dotransp))[::idir,::jdir]
#
# convert updated array back to ARCGIS raster
#
        tempras4 *= varscl
        tempras4 += meanrasval
        tempras4 += temprasSmooth
        output_raster = arcpy.NumPyArrayToRaster(tempras4, \
                          arcpy.Point(desc_ras.extent.XMin,desc_ras.extent.YMin), \
                          desc_ras.meanCellWidth,desc_ras.meanCellHeight)
#                             ,value_to_nodata=float(desc_ras.noDataValue))
        arcpy.DefineProjection_management(output_raster,desc_ras.spatialReference)
        s1="1 "+str(minrasval)+" "+str(maxrasval)+" "+str(meanrasval)+" "+str(stdrasval)
        s2="1 "+str(desc_ras.noDataValue)
        messages.addMessage("s1=["+s1+"]")
        messages.addMessage("s2=["+s2+"]")
        output_raster.save(parameters[2].valueAsText)
        arcpy.SetRasterProperties_management(parameters[2].valueAsText,statistics=s1,nodata=s2)
        output_layer_name = os.path.basename(parameters[2].valueAsText)+".lyr"
        arcpy.MakeRasterLayer_management(parameters[2].valueAsText,output_layer_name)
#           arcpy.ApplySymbologyFromLayer_management(output_layer_name,input_raster)
#           arcpy.CalculateStatistics_management(parameters[2].valueAsText,1,1,desc_ras.noDataValue)
#           output_raster.save(parameters[2].valueAsText)
#
        return
